%{
indexing

	description:
		"Scanner for mime messages."

	standards:
		"Based on RFC 822, 2045 and 2047."

	implementation:
		"1. Uses the specials from RFC 2047, so '/', '?' and '=' are%
		%   considered specials as well."

	author: "Berend de Boer"
	date: "$Date: $"
	revision: "$Revision: $"


class

	EPX_MIME_SCANNER

inherit

	YY_COMPRESSED_SCANNER_SKELETON
		export
			{EPX_MIME_PARSER} yyEnd_of_file_character
		end

	UT_CHARACTER_CODES
		export {NONE} all end

creation

	make

%}

%option outfile="epx_mime_scanner.e"
%option nodefault
%option case-insensitive
%option line


%x UNSTRUCTURED_FIELD_BODY STRUCTURED_FIELD_BODY
%x COMMENT DOMAIN_LITERAL ENCODED_TEXT NEXT_WORD QUOTED_STRING

---- definitions ----

CHAR  [\x00-\7f]
ALPHA [\101-\132]
DIGIT [0-9]
CTL   [\x00-\x1f\x7f]

%%
-------------------------------------------------------------------------------


-- White space
[ \t]+              -- Ignore white space

---- recognized fields ----

"Content-Disposition" {
							last_token := mime_tokens.FN_CONTENT_DISPOSITION
							to_condition := STRUCTURED_FIELD_BODY
							}
"Content-Length"     {
							last_token := mime_tokens.FN_CONTENT_LENGTH
							to_condition := STRUCTURED_FIELD_BODY
							}
"Content-Type"       {
							last_token := mime_tokens.FN_CONTENT_TYPE
							to_condition := STRUCTURED_FIELD_BODY
							}
"Date"               {
							last_token := mime_tokens.FN_DATE
							to_condition := STRUCTURED_FIELD_BODY
							}
"From"               {
							last_token := mime_tokens.FN_FROM
							to_condition := STRUCTURED_FIELD_BODY
							}
"Message-ID"         {
							last_token := mime_tokens.FN_MESSAGE_ID
							to_condition := STRUCTURED_FIELD_BODY
							}
"MIME-Version"       {
							last_token := mime_tokens.FN_VERSION
							to_condition := STRUCTURED_FIELD_BODY
							}
"Received"           {
							last_token := mime_tokens.FN_RECEIVED
							to_condition := STRUCTURED_FIELD_BODY
							}
"Resent-From"        {
							last_token := mime_tokens.FN_RESENT_FROM
							to_condition := STRUCTURED_FIELD_BODY
							}
"Resent-Reply-To"    {
							last_token := mime_tokens.FN_RESENT_REPLY_TO
							to_condition := STRUCTURED_FIELD_BODY
							}
"Resent-Sender"      {
							last_token := mime_tokens.FN_RESENT_SENDER
							to_condition := STRUCTURED_FIELD_BODY
							}
"Return-Path"        {
							last_token := mime_tokens.FN_RETURN_PATH
							to_condition := STRUCTURED_FIELD_BODY
							}
"Sender"             {
							last_token := mime_tokens.FN_SENDER
							to_condition := STRUCTURED_FIELD_BODY
							}
"To"                 {
							last_token := mime_tokens.FN_TO
							to_condition := STRUCTURED_FIELD_BODY
							}


---- all other fields ----

[^\x00-\x1f\:\x7f-\xff]+ {
							last_token := mime_tokens.FIELD_NAME
							last_value := text
							to_condition := UNSTRUCTURED_FIELD_BODY
							}

":"[ \t]+            { -- include space, so it's already skipped
							last_value := STRING_.make (64)
							last_token := Colon_code
							if to_condition /= 0 then
								set_start_condition (to_condition)
							end
							}

\r\n\r\n             |
\n\n                 {
							last_token := mime_tokens.NULL_LINE
							}


---- Structured field bodies (RFC 822 section 3.1.4)  ----

<STRUCTURED_FIELD_BODY>[ \t]+ { -- White space, ignore
							optional_start_keyword
							}
--<STRUCTURED_FIELD_BODY>"%\"    { } -- quoted-pair TODO
<STRUCTURED_FIELD_BODY>"=?"   { last_token := mime_tokens.ENCODED_WORD_START }
<STRUCTURED_FIELD_BODY>"<"    { last_token := Less_than_code }
<STRUCTURED_FIELD_BODY>">"    { last_token := Greater_than_code }
<STRUCTURED_FIELD_BODY>"@"    { last_token := At_code }
<STRUCTURED_FIELD_BODY>","    { last_token := Comma_code }
<STRUCTURED_FIELD_BODY>";"    { last_token := Semicolon_code }
<STRUCTURED_FIELD_BODY>":"    { last_token := Colon_code }
<STRUCTURED_FIELD_BODY>\\     { last_token := Backslash_code }
<STRUCTURED_FIELD_BODY>"."    { last_token := Dot_code }
<STRUCTURED_FIELD_BODY>"="    { last_token := Equal_code }
<STRUCTURED_FIELD_BODY>"/"    { last_token := Slash_code }
<STRUCTURED_FIELD_BODY>"?"    { last_token := Question_mark_code }
<STRUCTURED_FIELD_BODY>"("    {
							optional_start_keyword
							set_start_condition (COMMENT)
							}
<STRUCTURED_FIELD_BODY>"["    {
							last_value := STRING_.make (32)
							set_start_condition (DOMAIN_LITERAL)
							}
<STRUCTURED_FIELD_BODY>\" { --"
							last_value := STRING_.make (32)
							set_start_condition (QUOTED_STRING)
							}
<STRUCTURED_FIELD_BODY>\r?\n[ \t]+ { -- Folded-line white space, ignore
							optional_start_keyword
							}
<STRUCTURED_FIELD_BODY>\n\n     |
<STRUCTURED_FIELD_BODY>\r\n\r\n {
							last_token := mime_tokens.NULL_LINE
							stop_keyword
							set_start_condition (INITIAL)
							to_condition := 0
							}
<STRUCTURED_FIELD_BODY>\r?\n  {
							stop_keyword
							set_start_condition (INITIAL)
							to_condition := 0
							}
<STRUCTURED_FIELD_BODY>[^()<>@,;\:/\\\"\?\.=\[\]\x00-\x1f\x7f-\xff \t]+ { --"
							last_value := text
							if expect_keyword then
								if is_keyword (last_value) then
									last_token := last_keyword
								else
									last_token := mime_tokens.ATOM
								end
								temporarily_stop_keyword
							else
								if call_enable_keyword_after_space then
									enable_keyword_after_space
								end
								last_token := mime_tokens.ATOM
							end
							}
<STRUCTURED_FIELD_BODY>.      -- ignore all other stuff??


---- comments, possibly nested ----

<COMMENT>\\\(        { }
<COMMENT>\\\)        { }
<COMMENT>")"         {
							if nested_comment_count = 0 then
								set_start_condition (to_condition)
							else
								nested_comment_count := nested_comment_count - 1
							end
							}
<COMMENT>"("         { nested_comment_count := nested_comment_count + 1 }
<COMMENT>[^()]+      { }


---- domain literal ----
-- what to do when forbidden [ or CR is parsed? Now it's just appended.

<DOMAIN_LITERAL>\\. { last_value.append_character (text.item (2)) }
<DOMAIN_LITERAL>"]" {
							set_start_condition (to_condition)
							last_token := mime_tokens.DOMAIN_LITERAL
							}
<DOMAIN_LITERAL>\r?\n[ \t]+ {
							last_value.append_character (' ')
							}
<DOMAIN_LITERAL>[^]]+ { last_value.append_string (text) }

---- quoted-string ----
-- what to do when forbidden [ or CR is parsed? Now it's just appended.

<QUOTED_STRING>\\. { last_value.append_character (text.item (2)) }
<QUOTED_STRING>\" { --"
							set_start_condition (to_condition)
							last_token := mime_tokens.QUOTED_STRING
							}
<QUOTED_STRING>\r?\n[ \t]+ {
							last_value.append_character (' ')
							}
<QUOTED_STRING>[^"]+ { --"
							last_value.append_string (text)
							}


---- RFC 2047 B or Q encoding ----

<ENCODED_TEXT>"?="   { last_token := mime_tokens.ENCODED_WORD_STOP }
<ENCODED_TEXT>[^\x00-\x1f? \t]+ {
							last_value := text
							last_token := mime_tokens.ENCODED_TEXT
							}
<ENCODED_TEXT>.      { set_start_condition (to_condition) }


---- Unstructured field bodies (RFC section 3.1.3) ----

<UNSTRUCTURED_FIELD_BODY>"("      {
							set_start_condition (COMMENT) }
<UNSTRUCTURED_FIELD_BODY>[^\r\n]+ {
							last_value.append_string (text)
							}
<UNSTRUCTURED_FIELD_BODY>\r?\n[ \t]+ {
							last_value.append_character (' ')
							}
<UNSTRUCTURED_FIELD_BODY>\n\n     |
<UNSTRUCTURED_FIELD_BODY>\r\n\r\n {
							less (0)
							last_token := mime_tokens.FIELD_BODY
							set_start_condition (INITIAL)
							}
<UNSTRUCTURED_FIELD_BODY>\r?\n    {
							last_token := mime_tokens.FIELD_BODY
							set_start_condition (INITIAL)
							}
<UNSTRUCTURED_FIELD_BODY>\r       -- ignore lone CR's


----- error recovery -----

\r?\n               {
						  last_token := mime_tokens.CRLF
						  }

-------------------------------------------------------------------------------

<<EOF>>             { terminate }
.                   -- ignore all other charcters, i.e. ascii unprintables and high ascii

-------------------------------------------------------------------------------

%%

feature -- Last returned data

	last_value: STRING


feature -- Buffer setting

	new_stdc_file_buffer (a_file: STDC_TEXT_FILE): EPX_FILE_BUFFER is
			-- New input buffer for `a_file'.
		require
			file_not_void: a_file /= Void
			file_open: a_file.is_open
		do
			create Result.make (a_file)
		ensure
			new_buffer_not_void: Result /= Void
		end

	new_file_descriptor_buffer (a_fd: ABSTRACT_FILE_DESCRIPTOR): EPX_FILE_DESCRIPTOR_BUFFER is
			-- New input buffer for `a_fd'.
		require
			file_descriptor_not_void: a_fd /= Void
			file_descriptor_open: a_fd.is_open
		do
			create Result.make (a_fd)
		ensure
			new_buffer_not_void: Result /= Void
		end

feature -- Condition setting

	enable_keyword_after_space is
		do
			expect_keyword_after_space := True
		ensure
			no_keywords: not expect_keyword
		end

	optional_start_keyword is
			-- Start expect keyword if `expect_keyword_after_space'.
		do
			if expect_keyword_after_space  then
				start_keyword
			end
		ensure
			keyword_started:
				expect_keyword_after_space implies expect_keyword
		end

	reset_start_condition is
			-- Used in error recovery to return to start condition.
		do
			set_start_condition (INITIAL)
		end

	start_encoded_text is
			-- Next atom is encoded text.
		do
			set_start_condition (ENCODED_TEXT)
		end

	start_keyword is
			-- Treat certain atoms specially.
			-- The next atom will be tested if it is a keyword.
			-- The next atom will be treated as an atom, but the first
			-- space, folded-line or comment will turn keyword on at again,
			-- and so forth.
			-- This trick makes the scanner only slightly more state-full
			-- and the parser a lot simpler.
		do
			expect_keyword := True
		ensure
			expect_keywords: expect_keyword
		end

	stop_keyword is
			-- Stop treating certain atoms specially.
		do
			expect_keyword := False
			expect_keyword_after_space := False
			call_enable_keyword_after_space := False
		ensure
			no_more_keywords:
				not expect_keyword and
				not expect_keyword_after_space and
				not call_enable_keyword_after_space
		end

	temporarily_stop_keyword is
			-- Stop treating certain atoms specially for now, but set a
			-- flag so they will be enabled after the first space or
			-- comment.
		do
			expect_keyword := False
			expect_keyword_after_space := False
			call_enable_keyword_after_space := True
		ensure
			no_more_keywords_now:
				not expect_keyword and
				not expect_keyword_after_space
			keywords_later: call_enable_keyword_after_space
		end


feature {NONE} -- Mime specific tokens

	mime_tokens: expanded EPX_MIME_TOKENS

feature {NONE} -- State during parsing

	call_enable_keyword_after_space: BOOLEAN
			-- Flag to call `enable_keyword_after_space' when encountered an ATOM.

	expect_keyword: BOOLEAN
			-- Treat certain atoms specially.

	expect_keyword_after_space: BOOLEAN
			-- Treat certain atoms specially after a space, crlf in
			-- folded-line or comment has been ecountered.

	to_condition: INTEGER
			-- Used by ':' to switch to correct field body parser.

	nested_comment_count: INTEGER
			-- Keep track of nested comments.

	is_keyword (a_word: STRING): BOOLEAN is
			-- True if `a_word' is in `keywords'.
		local
			upper_word: STRING
		do
			upper_word := clone (a_word)
			upper_word.to_upper
			keywords.search (upper_word)
			Result := keywords.found
		end

	keywords: DS_HASH_TABLE [INTEGER, STRING] is
			-- List of atoms that can be treated as a keyword.
		once
			create Result.make (64)
			keywords.put (mime_tokens.KW_BY, "BY")
			keywords.put (mime_tokens.KW_FOR, "FOR")
			keywords.put (mime_tokens.KW_FROM, "FROM")
			keywords.put (mime_tokens.KW_ID, "ID")
			keywords.put (mime_tokens.KW_MULTIPART, "MULTIPART")
			keywords.put (mime_tokens.KW_VIA, "VIA")
			keywords.put (mime_tokens.KW_WITH, "WITH")

			keywords.put (mime_tokens.KW_SUN, "SUN")
			keywords.put (mime_tokens.KW_MON, "MON")
			keywords.put (mime_tokens.KW_TUE, "TUE")
			keywords.put (mime_tokens.KW_WED, "WED")
			keywords.put (mime_tokens.KW_THU, "THU")
			keywords.put (mime_tokens.KW_FRI, "FRI")
			keywords.put (mime_tokens.KW_SAT, "SAT")
		ensure
			result_not_void: Result /= Void
		end

	last_keyword: INTEGER is
			-- Last keyword if `is_keyword' returned True.
		do
			Result := keywords.found_item
		end


invariant

	never_a_keyword_in_start_condition:
		start_condition = INITIAL implies
			(not expect_keyword and not expect_keyword_after_space)

end
