indexing

	description:

		"Base class for other classes that create or get assigned a handle%
		%to an external resource like a file, socket or memory block. %
		%This class can become owner of the resource in which case it%
		%automatically closes the resource when garbage collected. It also%
		%contains protection against misuse of resources by checking against%
		%set limits."

	author: "Berend de Boer"
	date: "$Date: 2003/01/09 $"
	revision: "$Revision: #1 $"


deferred class

	STDC_HANDLE [H]


inherit

	STDC_BASE

	MEMORY
		export
			{NONE} all
		redefine
			dispose
		end


feature {NONE} -- Resources retrieved from other applications

	attach_to_handle (a_handle: H; a_become_owner: BOOLEAN) is
			-- This class will give access to `a_handle'. It will not
			-- become the owner of it, so it will not be automatically
			-- closed, unless `become_owner' is called.
		require
			closed: not is_open
			is_handle: a_handle /= unassigned_value
			capacity_set: capacity > 0
			resource_usage_within_limits: resource_usage_can_be_increased
		do
			set_handle (a_handle, a_become_owner)
		ensure
			open: is_open
			owner_set: a_become_owner = is_owner
			resource_accounted_for:
				(a_become_owner implies
					resource_count = old resource_count + capacity) and then
				(not a_become_owner implies
					resource_count = old resource_count)
		end


feature -- Access

	is_open: BOOLEAN is
		do
			Result := handle /= unassigned_value
		end

	is_owner: BOOLEAN
			-- Does this object close the stream on `close' or `dispose'?
			-- Only for resources that are owned, are resource limits checked.

	resource_usage_can_be_increased: BOOLEAN is
			-- Can the number of allocated resources increased with `capacity'?
		require
			capacity_set: capacity > 0
		deferred
		end


feature -- Influence ownership of the handle. Can help to influence subtile garbage collector problems

	become_owner is
			-- This class will own its handle. This is the only function
			-- that actually increases the resource count.
		require
			open: is_open
			not_owner: not is_owner
			resource_usage_within_limits: resource_usage_can_be_increased
		do
			is_owner := True
			increase_resource_count
		ensure
			owner: is_owner
			open_resources_increased:
				resource_count = old resource_count + capacity
		end

	unown is
			-- Resource will not be closed on dispose. Calling close will
			-- be forbidden.
		require
			owner: is_owner
		do
			decrease_resource_count
			is_owner := False
		ensure
			not_owner: not is_owner
			open_resources_decreased:
				resource_count = old resource_count - capacity
		end


feature -- Close

	close is
			-- Close the resource.
		require
			open: is_open
			owner: is_owner
		local
			success: BOOLEAN
		do
			success := do_close
			capacity := 0
			if not success then
				raise_posix_error
			end
		ensure
			closed: not is_open
			not_owner: not is_owner
			resource_accounted_for:
				(old is_owner implies
					resource_count = old resource_count - old capacity) and then
				(not old is_owner implies
					resource_count = old resource_count)
		end

	detach is
			-- Forget the resource. Resource is not closed.
		require
			open: is_open
		do
			clear_handle
		ensure
			closed: not is_open
			not_owner: not is_owner
			resource_accounted_for:
				(old is_owner implies
					resource_count = old resource_count - old capacity) and then
				(not old is_owner implies
					resource_count = old resource_count)
		end


feature -- Cleanup

	dispose is
			-- Close handle if owner.
		local
			success: BOOLEAN
		do
			if is_open and then is_owner then
				success := do_close
			end
		end


feature -- Resource

	capacity: INTEGER
			-- Number of resources that are in use by `handle'. For a
			-- file this is 1, for a memory handle, this is the number of
			-- bytes.

	handle: H
			-- Identifier of resource tracked by this class.


feature {NONE} -- Low level handle functions

	clear_handle is
			-- Give handle its default value. This is the only function
			-- that may do so, else resource counting may be
			-- incorrect. This routine may never call another object,
			-- else it cannot be used safely in `dispose'.
		do
			if is_owner then
				unown
			end
			handle := unassigned_value
			capacity := 0
		ensure
			handle_unassigned: handle = unassigned_value
			resource_accounted_for:
				(old is_owner implies
					resource_count = old resource_count - old capacity) and then
				(not old is_owner implies
					resource_count = old resource_count)
		end

	do_close: BOOLEAN is
			-- Close resource. Return False if an error occurred. Error
			-- value should be in `errno'. This routine may never call
			-- another object, else it cannot be used safely in
			-- `dispose'.
			-- This routine is usely redefined to actually close or
			-- deallocate the resource in addition of resetting `handle'.
		require
			open: is_open
			owner: is_owner
		do
			clear_handle
		ensure
			closed: not is_open
			resource_accounted_for:
				resource_count = old resource_count - old capacity
		end

	set_handle (a_handle: H; a_become_owner: BOOLEAN) is
			-- Only function that sets `handle'. Should only be called
			-- with a valid handle, call `clear_handle' to reset the
			-- handle.
		require
			is_handle: a_handle /= unassigned_value
			handle_not_set: handle = unassigned_value
			capacity_set: capacity > 0
			unowner: not is_owner
			resource_usage_within_limits: a_become_owner implies resource_usage_can_be_increased
		do
			handle := a_handle
			if a_become_owner then
				become_owner
			end
		ensure
			handle_set: handle = a_handle
			owner_set: a_become_owner = is_owner
			resource_accounted_for:
				(a_become_owner implies
					resource_count = old resource_count + capacity) and then
				(not a_become_owner implies
					resource_count = old resource_count)
		end

	unassigned_value: H is
			-- The value that indicates that `handle' is unassigned.
		deferred
		end


feature {NONE} -- Counting of allocated resource

	decrease_resource_count is
			-- Decrease number of allocated resource by `capacity'.
			-- Due to limitations of certain Eiffel implementations, this
			-- routine may not call any other object. Calling C code is safe.
		require
			capacity_set: capacity > 0
		deferred
		ensure
			open_resources_decreased:
				resource_count = old resource_count - capacity
		end

	increase_resource_count is
			-- Increase number of allocated resources by `capacity'.
		require
			capacity_set: capacity > 0
			resource_usage_within_limits: resource_usage_can_be_increased
		deferred
		ensure
			open_resources_increased:
				resource_count = old resource_count + capacity
		end

	resource_count: INTEGER is
			-- Currently allocated number of resources. It's a global
			-- number, counting the `capacity' of all owned resources of
			-- this type.
		deferred
		end


invariant

	capacity_not_negative: capacity >= 0
	valid_capacity: is_open = (capacity > 0)
	open_implies_handle_assigned: is_open = (handle /= unassigned_value)
	owned_implies_open: is_owner implies is_open
	owned_implies_handle_assigned: is_owner implies handle /= unassigned_value

end
